function [ total_reward,steps,Q,Model ] = Episode( maxsteps,Q,Model,alpha,gamma,epsilon,statelist,actionlist,grafic,maze,start,goal,p_steps )
% Episode do one episode of reaching the goal position in maze table
% maxstepts: the maximum number of steps per episode
% Q: the current QTable
% alpha: the current learning rate
% Model:
% gamma: the current discount factor
% epsilon: probablity of a random action
% statelist: the list of states
% actionlist: the list of actions
% grafic: whether display the maze map in each round
% maze: maze map
% start: start position
% goal: goal position
% p_steps" planning steps


%total_reward: total rewards
%steps: number of steps to reach the goal position


x = start; % x is the current position
steps = 0;
total_reward = 0;


% find the current position in the state list
[~, s] = min(dist(statelist,x'));

% selects an action using the epsilon greedy selection strategy
a = greedy_selection(Q,s,epsilon);


for i=1:maxsteps    
    % flag indicates whether move to the target position
    flag = 0;
    
    % convert the index of the action into an action value, actually it is
    % same as action = a;
    action = actionlist(a);    
    
    %do the selected action and move to the next state;    
    xp  = DoAction(action, x, maze);    
    
    % observe the reward at state xp and the final state flag
    if xp == goal
        flag = 1;
        r = 1;
    else
        r = 0;
    end
    
    total_reward = total_reward + r;
    
    % find the current position in the state list
    [~, sp] = min(dist(statelist,xp'));
    
    
    % selects an action using the epsilon greedy selection strategy
    ap = greedy_selection(Q,sp,epsilon);
    
    
    % Update the Qtable for the previous step before the action a in Q
    % table
    Q = UpdateQLearning(s, a, r, sp, Q, alpha, gamma);
    
    % keep the next state and reward after action a
    Model(s,a,1) = sp;
    Model(s,a,2) = r;
    
    % planning
    Q = RandomPlanning(Q, Model, p_steps, alpha, gamma);
    
    
    %update the current variables
    s = sp;
    a = ap;
    x = xp;
    
    
    %increment the step counter.
    steps=steps+1;
    
   
    % Plot one move in maze map
    if (grafic==true)
        Plot(x,maze,start,goal,['PLANNING (N=' num2str(p_steps) ')']);
    end
    
    % if reach to the target position, breaks the episode
    if (flag == 1)
        break
    end
    
end